class Guild:
  def __init__(self, guild_data, client):
    self.id = guild_data.get("id", None)
    self.name = guild_data.get("name", None)
    self.members = guild_data.get("members", None)
    self.channels = guild_data.get("channels", None)
    self.member_count = guild_data.get("member_count", None)
    #self.categories = guild_data["categories"]


from .channel import Channel
from .user import User
from .guild import Guild

class Message:
  def __init__(self, message_data, client):
    self.id = message_data['id']
    self._JSON = message_data
    self.guild_id = message_data["guild_id"]
    self.content = message_data.get('content')
    self.author = User(message_data.get('author', message_data.get('member').get('user')), client)
    self.channel = Channel(message_data['channel_id'], client)

  async def remove(self):
    return await self.channel.remove_message(self.id)
  
  async def reply(self, content=None, embed=None):
    return await self.channel.send(content, embed, reference=self.id)


def Red():
    """Returns the hex code for the color red."""
    return 0xFF0000

def Green():
    """Returns the hex code for the color green."""
    return 0x00FF00

def Blue():
    """Returns the hex code for the color blue."""
    return 0x0000FF

def Purple():
    """Returns the hex code for the color purple."""
    return 0x800080

def Yellow():
    """Returns the hex code for the color yellow."""
    return 0xFFFF00

def Orange():
    """Returns the hex code for the color orange."""
    return 0xFFA500

def Black():
    """Returns the hex code for the color black."""
    return 0x000000

def White():
    """Returns the hex code for the color white."""
    return 0xFFFFFF

def Gray():
    """Returns the hex code for the color gray."""
    return 0x808080


class Embed:
  def __init__(self, title="", description="", color=0):
    self.title = title
    self.description = description
    self.color = color
    self.fields = []
    self.footer_text = None
    self.footer_icon = None
    self.author_name = None
    self.author_url = None
    self.author_icon = None

  def add_field(self, name, value, inline=False):
    self.fields.append({"name": name, "value": value, "inline": inline})

  def set_footer(self, text, icon_url=None):
    self.footer_text = text
    self.footer_icon = icon_url

  def set_author(self, author, url=None, icon_url=None):
    self.author_name = author
    self.author_url = url
    self.author_icon = icon_url

  def to_dict(self):
    return {"title": self.title, "description": self.description, "color": self.color, "author": {"name": self.author_name, "url": self.author_url, "icon_url": self.author_icon}, "footer": {"text": self.footer_text, "icon_url": self.footer_icon}, "fields": self.fields}


class User:
  def __init__(self, user_data, client):
    self.id = user_data["id"]
    self.username = user_data['username']
    self.display = user_data.get("global_name")
    self.bot = user_data.get('bot', False)
    self._client = client
    self.dm = None
    self.session = client.http.session

  def set_dm(self, id: int):
    self.dm = id

  @property
  def mention(self) -> str:
    return f"<@{self.id}>"

  async def remove_message(self, message_id):
    if self.dm == None:
      raise Exception("You haven't created a DM before trying to send something to user")
    url = f'https://discord.com/api/v10/channels/{self.dm}/messages/{message_id}'
    headers = {
      'Authorization': f'Bot {self._client.token}',
      'Content-Type': 'application/json',
    }
    async with self.session.delete(url, headers=headers) as response:
      return await response.json()
  
  async def create_dm(self):
    url = f'https://discord.com/api/v10/users/@me/channels'
    headers = {
      'Authorization': f'Bot {self._client.token}',
      'Content-Type': 'application/json',
    }
    payload = {'recipient_id': self.id}
    async with self.session.post(url, headers=headers, json=payload) as response:
      if response.status_code == 200:
        idd = await response.json().get('id')
        self.set_dm(idd)
      else:
        raise Exception(f"Failed to create DM channel: {dm_response.json()}")
  
  async def send(self, content=None, embed=None, reference=None, embeds=None):
    if self.dm == None:
      raise Exception("You haven't created a DM before trying to send something to user")
    url = f'https://discord.com/api/v10/channels/{self.dm}/messages'
    headers = {
      'Authorization': f'Bot {self._client.token}',
      'Content-Type': 'application/json',
    }
    actual_embeds = []
    if embed:
      actual_embeds.append(embed)
    if embeds:
      actual_embeds.extend(embeds)
                            
    payload = {
      'content': str(content) if content else None,
      'embeds': [e.to_dict() for e in actual_embeds if hasattr(e, 'to_dict')],
      'message_reference': {'message_id': reference} if reference else None
    }
    async with self.session.post(url, headers=headers, json=payload) as response:
      return await response.json()


import asyncio, time, aiohttp, logging, json, websockets, inspect
from typing import Any, Callable, Dict, List, Optional, Union
from .message import Message
from .msg import Msg
from . import intents
from .intents import all
from .embed import Embed
from .guild import Guild
from .user import User
from . import color as Color
from .http import HTTPClient
from enum import IntEnum

version = "1.0.0"

class ApplicationCommandOptionType(IntEnum):
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10
    ATTACHMENT = 11

class Command:
  def __init__(self, func, name, **attrs):
    self.func = func
    self.name = name or func.__name__
    self.params = attrs.get('params', {})
    self.description = attrs.get('description', None)
    self.slash = attrs.get('slash', False)
    self.guild_ids = attrs.get('guild_ids', [])
    self.hybrid = attrs.get('hybrid', False)
    self.aliases = attrs.get('aliases', [])
    self.help_text = attrs.get('help', None)

  async def invoke(self, context, *args, **kwargs):
    return await self.func(context, *args, **kwargs)

  def _get_param_type(self, annotation: Any) -> Optional[ApplicationCommandOptionType]:
    if annotation == str:
      return ApplicationCommandOptionType.STRING
    elif annotation == int:
      return ApplicationCommandOptionType.INTEGER
    elif annotation == bool:
      return ApplicationCommandOptionType.BOOLEAN
    elif annotation == float:
      return ApplicationCommandOptionType.NUMBER
    elif annotation == User:
      return ApplicationCommandOptionType.USER
    return None

  def to_application_command_payload(self) -> Optional[Dict[str, Any]]:
    """Converts this command into a Discord Application Command JSON payload."""
    if not self.slash:
      if not self.hybrid:
        return None

    payload_options: List[Dict[str, Any]] = []
    param_names = list(self.params.keys())
    if param_names and (param_names[0] == 'ctx' or param_names[0] == 'self'):
      actual_params = list(self.params.values())[1:]
    else:
      actual_params = list(self.params.values())

    for param in actual_params:
      if param.annotation == inspect.Parameter.empty:
        option_type = ApplicationCommandOptionType.STRING
        param_description = f"The {param.name} argument."
      else:
        option_type = self._get_param_type(param.annotation)
        if option_type is None:
          print(f"Warning: Could not map type {param.annotation} for param '{param.name}' in command '{self.name}'. Skipping.")
          continue
        param_description = f"Value for {param.name}."
        is_required = (param.default == inspect.Parameter.empty)

        payload_options.append({
          "name": param.name,
          "description": param_description,
          "type": option_type.value,
          "required": is_required,
          # "choices": [] # You'd handle choices here if supported
        })

    return {
      "name": self.name,
      "description": self.description if self.description else "No description provided.",
      "options": payload_options,
      # "type": 1, # CHAT_INPUT, default
      # "dm_permission": True, # Default
      # "default_member_permissions": None # For permission restricting
    }

class Client:
  def __init__(self, token, intents=None, prefix=None, builtin_help=True):
    self.token = token
    self.prefix = prefix
    self.event_handlers = {}
    self.commands = {}
    self.guilds = []
    self.user = None
    self.id = None
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    logger = logging.getLogger(__name__)
    self.logger = logger
    self.http = HTTPClient(token=token, client=self)
    if intents == None:
      self.intents = all()
      self.logger.warn("Intents were not set! Using default Intents (all)")
    else:
      self.intents = intents
    if prefix == None:
      self.logger.warn("Prefix was not set! Bot commands won't be available!")
    if builtin_help:
      async def help_cmd(msg, page=0):
        help_embed = Embed(title="Help GUI", description=f"Commands Are specified down below \n **Page {page}**")
        fro = page*10 if not page*10 > len(self.commands) else int(len(self.commands)/10)
        to = page*10+10 if not page*10+10 > len(self.commands) else len(self.commands)
        commands_list = list(self.commands.keys())[fro:to]
        for command in commands_list:
          command = self.commands[command]
          help_embed.add_field(name=f"{command.name}", value=f"**Description:** {command.description} \n **Help:** {command.help_text}")
        global version
        help_embed.set_footer(text=f"PaWpD {version}")
        await msg.send(embed=help_embed)
      help_command = Command(help_cmd, "help")
      self.add_command(help_command)
    
  def add_command(self, command: Command):
    if command.name in self.commands:
      self.logger.info(f"Warning: Command '{command.name}' is being overwritten.")
    self.commands[command.name] = command

  async def sync(self, *, guild_id: Optional[int] = None) -> List[Dict[str, Any]]:
    """
        Registers or updates application commands with Discord.

        Args:
            guild_id: If provided, syncs commands to this specific guild.
                      Otherwise, syncs globally.

        Returns:
            A list of the application command data as returned by Discord.
    """
    if not self.id:
      self.logger.error("Error: Bot id is not set. Cannot sync commands.")
      return []

    payloads_to_send: List[Dict[str, Any]] = []
    for cmd_name, command_obj in self.commands.items():
      if command_obj.slash == False:
        continue
      if command_obj.hybrid == False:
        continue 
      is_guild_specific_command_match = guild_id and command_obj.guild_ids and guild_id in command_obj.guild_ids
      is_global_command_for_global_sync = not guild_id and not command_obj.guild_ids
      is_global_command_for_guild_sync = guild_id and not command_obj.guild_ids

      should_include = False
      if guild_id:
        if command_obj.guild_ids:
          if guild_id in command_obj.guild_ids:
            should_include = True
        else:
          should_include = True
      else:
        if not command_obj.guild_ids:
          should_include = True

      if should_include:
        payload = command_obj.to_application_command_payload()
        if payload:
          payloads_to_send.append(payload)

    if not payloads_to_send:
      self.logger.info(f"No application commands found to sync for {'guild ' + str(guild_id) if guild_id else 'global'}.")
      return []

    try:
      if guild_id:
        self.logger.info(f"Syncing {len(payloads_to_send)} commands to guild {guild_id}...")
                
        response_data = await self.http.bulk_override_guild_commands(
          self.id, guild_id, payloads_to_send
        )
        self.logger.info(f"Successfully synced {len(response_data)} commands to guild {guild_id}.")
      else:
        self.logger.info(f"Syncing {len(payloads_to_send)} commands globally...")
  
        response_data = await self.http.bulk_override_global_commands(
          self.id, payloads_to_send
        )
        self.logger.info(f"Successfully synced {len(response_data)} commands globally.")
        return response_data
    except Exception as e:
      self.logger.error(f"Error during command sync: {e}")
      return []

  def event(self, coro):
    if asyncio.iscoroutinefunction(coro):
      self.event_handlers[coro.__name__] = coro
    else:
      raise TypeError('Event handler must be a coroutine function')
    return coro

  def command(self, name=None, **attrs):
    def decorator(func):
      cmd_name = name or func.__name__
      command_obj = Command(func, cmd_name, **attrs)
      self.commands[command_obj.name] = command_obj
      for alias in command_obj.aliases:
        self.commands[alias] = command_obj
      return func
    return decorator

  async def latency(self):
    start_time = time.perf_counter()
    await self.http.request(method="GET", path="/gateway")
    latency = (time.perf_counter() - start_time) * 1000  
    return latency
        
  async def _connect(self):
    while True:
      try:
        self.logger.info("Starting Connection to Discord API")
        async with websockets.connect("wss://gateway.discord.gg/?v=10&encoding=json", max_size=10000000) as ws:
          await self._identify(ws)
          hello_message = await ws.recv()
          hello_data = json.loads(hello_message)
          heartbeat_interval = hello_data['d']['heartbeat_interval'] / 1000
          async def send_heartbeat():
            while True:
              await asyncio.sleep(heartbeat_interval)
              heartbeat_payload = {
                'op': 1,
                'd': None
              }
              await ws.send(json.dumps(heartbeat_payload))
          asyncio.create_task(send_heartbeat())                 
                    
          self.logger.info(f"Connection Successful. Latency {await self.latency()} MS")
          async for message in ws:
            try:
              await self._handle(message)
            except websockets.exceptions.PayloadTooBig:
              self.logger.error("Payload too big. Skipping this message.")
            except Exception as e:
              self.logger.error(e)

      except websockets.ConnectionClosed as e:
        self.logger.error(f'Connection closed: {e}')
        await asyncio.sleep(1)
        
      except Exception as e:
        self.logger.error(f'An error occurred: {e}')
        await asyncio.sleep(1)
      self.logger.info("Trying to Establish Connection")

  async def _identify(self, ws):
    await ws.send(json.dumps({
      "op": 2,
      "d": {
        "token": self.token,
        'intents': self.intents,
        "properties": {
          "$os": "PaWpD",
          "$browser": "PaWpD",
          "$device": "PaWpD",
        },
        "presence": {"status": "online", "afk": False},
      },
    }))

  async def _handle(self, message):
    event = json.loads(message)
    event_name = event.get('t')
    event_data = event.get('d')
    if event_name == 'MESSAGE_CREATE':
      handler = self.event_handlers.get('on_message_create')
      msg_obj = Message(event_data, self)
      if handler:
        await handler(msg_obj)
      elif self.prefix and event_data['content'].startswith(self.prefix):
        await self.process(event_data, _side=True)
    elif event_name == "INTERACTION_CREATE":
      name = event_data['data']['name']
      handler = self.commands.get(name)
      args = event_data['data'].get('options')
      message_obj = Msg(event_data, self, interaction=True)
      if handler:
        if args:
          await handler.invoke(message_obj, *args)
        else:
          await handler.invoke(message_obj)
    elif event_name == "READY":
      handler = self.event_handlers.get('on_ready')
      self.user = User(event_data['user'], self)
      self.id = self.user.id
      for guild in event_data['guilds']:
        self.guilds.append(Guild(guild, self))
      if handler:         
        await handler(self.user)
    elif event_name == 'MESSAGE_UPDATE':
      handler = self.event_handlers.get('on_message_update')
      msg_obj = Msg(event_data, self)
      if handler:
        await handler(msg_obj)
    elif event_name == 'MESSAGE_DELETE':
      handler = self.event_handlers.get('on_message_delete')
      msg_obj = Message(event_data, self)
      if handler:
        await handler(msg_obj)
    elif event_name == "GUILD_CREATE":
      handler = self.event_handlers.get('on_guild_join')
      guild_obj = Guild(event_data, self)
      self.guilds.append(guild_obj)
      if handler:
        await handler(guild_obj)
    elif event_name == "GUILD_DELETE":
      handler = self.event_handlers.get('on_guild_leave')
      guild_obj = Guild(event_data, self)
      self.guilds.pop(guild_obj)
      if handler:
        await handler(guild_obj)
    elif event_name == "GUILD_UPDATE":
      handler = self.event_handlers.get('on_guild_update')
      guild_obj = Guild(event_data, self)
      if handler:
        await handler(guild_obj)
    elif event_name == "GUILD_MEMBER_ADD":
      handler = self.event_handlers.get('on_member_join')
      user_obj = User(event_data, self)
      if handler:
        await handler(user_obj)
    elif event_name == "GUILD_MEMBER_REMOVE":
      handler = self.event_handlers.get('on_member_leave')
      user_obj = User(event_data, self)
      if handler:
        await handler(user_obj)

  async def process(self, event_data, _side=False):
    if _side == False:
      event_data = event_data._JSON
    parts = event_data['content'][len(self.prefix):].split()
    command = parts[0]
    args = parts[1:]          
    msg_obj = Msg(event_data, self)
    handler = self.commands.get(command).invoke
    if handler:
      await handler(msg_obj, *args)
        
  def start(self):
    asyncio.get_event_loop().run_until_complete(self.http._get_session())
    asyncio.get_event_loop().run_until_complete(self._connect())
    asyncio.get_event_loop().run_forever()


class Channel:
  def __init__(self, channel_id, client):
    self.id = channel_id
    self._client = client
    self.session = client.http.session

  async def remove_message(self, message_id):
    url = f'https://discord.com/api/v10/channels/{self.id}/messages/{message_id}'
    headers = {
      'Authorization': f'Bot {self._client.token}',
      'Content-Type': 'application/json',
    }
    async with self.session.delete(url, headers=headers) as response:
      return await response.json()
  
  async def send(self, content=None, embed=None, reference=None, embeds=None, ephemeral=False, interaction_token=None):     
    url = f'https://discord.com/api/v10/channels/{self.id}/messages'
    headers = {
      'Authorization': f'Bot {self._client.token}',
      'Content-Type': 'application/json',
    }
    actual_embeds = []
    if embed:
      actual_embeds.append(embed)
    if embeds:
      actual_embeds.extend(embeds)
    payload = {
      'content': str(content) if content else None,
      "embeds": [e.to_dict() for e in actual_embeds if hasattr(e, 'to_dict')],
      'message_reference': {'message_id': reference} if reference else None
    }
    if ephemeral:
      payload['flags'] = 64
    if reference:
      message_data = await self._client.http.create_followup_message(
        application_id=_client.id,
        interaction_token=interaction_token,
        data=payload
      )
      if message_data:
        return message_data
    async with self.session.post(url, headers=headers, json=payload) as response:
      return await response.json()


from typing import Any, Callable, Dict, List, Optional, Union
import aiohttp

class HTTPClient:
  def __init__(self, token: str, client):
    self.client = client
    self.token = token
    self.session = None

  async def _get_session(self):
    self.session = aiohttp.ClientSession()
    self.client.logger.info("Aiohttp Session started")

  async def _close_session(self):
    await self.session.close()
    self.client.logger.info("Aiohttp Session stopped")

  async def request(self, method: str, path: str, json_payload: Optional[Dict] = None) -> Any:
    path = f"https://discord.com/api/v10{path}"
    if method == "GET":
      async with self.session.get(path, headers={'Authorization': f'Bot {self.token}', 'Content-Type': 'application/json'}) as response:
        data = await response.json()
    elif method == "POST":
      async with self.session.post(path, headers={'Authorization': f'Bot {self.token}', 'Content-Type': 'application/json'}, json=json_payload) as response:
        data = await response.json()
    elif method == "DELETE":
      async with self.session.delete(path, headers={'Authorization': f'Bot {self.token}', 'Content-Type': 'application/json'}) as response:
        data = await response.json()
    elif method == "PUT":
      async with self.session.put(path, headers={'Authorization': f'Bot {self.token}', 'Content-Type': 'application/json'}, json=json_payload) as response:
        data = await response.json()
    else:
      return []
    if response.status == 200:
      return data
    if response.sttaus == 204:
      return
    else:
      raise Exception(f"HTTP Error: {response.status}")

  async def bulk_override_global_commands(self, application_id: str, commands_payload: List[Dict]) -> List[Dict]:
    return await self.request("PUT", f"/applications/{application_id}/commands", json_payload=commands_payload)

  async def create_interaction_response(self, interaction_id: str, interaction_token: str, response_type: int, data: Optional[Dict]):
    return await self.request('POST', f'/interactions/{interaction_id}/{interaction_token}/callback', json_payload=data)

  async def create_followup_message(self, application_id: str, interaction_token: str, data: Dict) -> Dict:
    return await self.request('POST', f'/webhooks/{application_id}/{interaction_token}', json_payload=data)

  async def bulk_override_guild_commands(self, application_id: str, guild_id: int, commands_payload: List[Dict]) -> List[Dict]:
    return await self.request("PUT", f"/applications/{application_id}/guilds/{guild_id}/commands", json_payload=commands_payload)


from .channel import Channel
from .user import User
from .message import Message

class Msg:
  def __init__(self, message_data, client, interaction=False):
    self.id = message_data['id']
    self._JSON = message_data
    self.content = message_data.get('content', None)
    self.message = Message(message_data, client)
    self.author = User(message_data.get('author', message_data.get('member').get('user')), client)
    self.channel = Channel(message_data['channel_id'], client)
    self._client = client
    if interaction:
      self._deferred = False
      self._responded = False
      self._interaction = interaction
      self._token: str = message_data['token']
      self.version: int = message_data['version']

  async def defer(self, *, ephemeral: bool = False, thinking: bool = True):
    if self._responded or self._deferred:
      return
    response_type = 5
    data = {}
    if ephemeral:
      data['flags'] = 64
    await self._client.http.create_interaction_response(
      interaction_id=self.interaction.id,
      interaction_token=self.interaction.token,
      response_type=response_type,
      data=data if data else None
    )
    self._deferred = True
    self._responded = True

  async def remove(self):
    return await self.channel.remove_message(self.id)
    
  async def send(self, content=None, embed=None, embeds=None, ephemeral=False):
    return await self.channel.send(content=content, embed=embed, embeds=embeds, ephemeral=ephemeral)

  async def reply(self, content=None, embed=None, embeds=None, ephemeral=False):
    if _interaction:
      self._responded = True
    return await self.channel.send(content=content, embed=embed, embeds=embeds, reference=self.id, ephemeral=False, interaction_token=_token)